#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "rti_me_c.h"
#include "disc_dpde/disc_dpde_discovery_plugin.h"
#include "HelloWorldApplication.h"
#include "HelloWorld.h"
#include "wh_sm/wh_sm_history.h"
#include "rh_sm/rh_sm_history.h"
#include "netio/netio_udp.h"

#ifdef __ANDROID__
#include <android/log.h>
#include <android_native_app_glue.h>
#include <EGL/egl.h>
#include <GLES/gl.h>
#endif

void
Application_help(char *appname)
{
    printf("%s [options]\n", appname);
    printf("options:\n");
    printf("-h                 - This text\n");
    printf("-domain <id>       - DomainId (default: 0)\n");
    printf("-udp_intf <intf>   - udp interface (no default)\n");
    printf("-peer <address>    - peer address (no default)\n");
    printf("-count <count>     - count (default -1)\n");
    printf("-sleep <ms>        - sleep between sends (default 1s)\n");
    printf("\n");
}

void
Application_printf(const char *format,...)
{
    va_list arglist;
    va_start(arglist, format);

#ifdef __ANDROID__
    (void)__android_log_vprint(ANDROID_LOG_DEBUG,"com.rti",format,arglist);
#else
    vprintf(format,arglist);
#endif
    va_end(arglist);
}

struct Application*
Application_create(struct android_app* android_state,
                   const char *local_participant_name,
                   const char *remote_participant_name,
                   DDS_Long domain_id, char *udp_intf, char *peer,
                   DDS_Long sleep_time, DDS_Long count)
{
    DDS_ReturnCode_t retcode;
    DDS_DomainParticipantFactory *factory = NULL;
    struct DDS_DomainParticipantFactoryQos dpf_qos =
        DDS_DomainParticipantFactoryQos_INITIALIZER;
    struct DDS_DomainParticipantQos dp_qos =
        DDS_DomainParticipantQos_INITIALIZER;
    struct DPDE_DiscoveryPluginProperty discovery_plugin_properties =
        DPDE_DiscoveryPluginProperty_INITIALIZER;
    DDS_Boolean success = DDS_BOOLEAN_FALSE;
    struct Application *application = NULL;
    RT_Registry_T *registry = NULL;
    struct UDP_InterfaceFactoryProperty *udp_property = NULL;

/* uncomment to increase verbosity level 
	OSAPILog_set_verbosity(OSAPI_LOG_VERBOSITY_WARNING);*/
    application = (struct Application *)malloc(sizeof(struct Application));

    if (application == NULL)
    {
        Application_printf("failed to allocate application\n");
        goto done;
    }

    memset(application, 0, sizeof(application));
    application->exit_app = RTI_FALSE;
#ifdef __ANDROID__
    application->wait_for_ready = RTI_TRUE;
#else
    application->wait_for_ready = RTI_FALSE;
#endif

    application->android_state = android_state;

    application->ready_sem = OSAPI_Semaphore_new();
    application->data_sem = OSAPI_Semaphore_new();

    application->sleep_time = sleep_time;
    application->count = count;

    factory = DDS_DomainParticipantFactory_get_instance();

    registry =
        DDS_DomainParticipantFactory_get_registry
        (DDS_DomainParticipantFactory_get_instance());

    RT_Registry_register(registry, "wh",
            WHSM_HistoryFactory_get_interface(), NULL, NULL);

    RT_Registry_register(registry, "rh",
            RHSM_HistoryFactory_get_interface(), NULL, NULL);

    /* Configure UDP transport's allowed interfaces */
    if (!RT_Registry_unregister(registry, "_udp", NULL, NULL))
    {
        Application_printf("failed to unregister udp\n");
        goto done;
    }

    udp_property = (struct UDP_InterfaceFactoryProperty *)
                       malloc(sizeof(struct UDP_InterfaceFactoryProperty));
    *udp_property = UDP_INTERFACE_FACTORY_PROPERTY_DEFAULT;

    /* Architectures with default loopback interfaces */
    REDA_StringSeq_set_maximum(&udp_property->allow_interface,1);
    REDA_StringSeq_set_length(&udp_property->allow_interface,1);

    if (udp_intf != NULL)
    { /* use interface supplied on command line */
        *REDA_StringSeq_get_reference(&udp_property->allow_interface,0) = 
            DDS_String_dup(udp_intf);
    }
    else
    {                /* use hardcoded interface */
#if defined(RTI_DARWIN)
    *REDA_StringSeq_get_reference(&udp_property->allow_interface,0) = "en0";
#elif defined (RTI_LINUX)
#if defined(__ANDROID__) && !defined(ANDROID_EMU)
    *REDA_StringSeq_get_reference(&udp_property->allow_interface,0) = "wlan0";
#else
    *REDA_StringSeq_get_reference(&udp_property->allow_interface,0) = "eth0";
#endif
#elif defined (RTI_VXWORKS)
        *REDA_StringSeq_get_reference(&udp_property->allow_interface,0) = "geisc0";
#else
        *REDA_StringSeq_get_reference(&udp_property->allow_interface,0) = "ce0";
#endif
    }


#if defined(ANDROID_EMU)
    Application_printf("Configure NAT");

    UDP_NatEntrySeq_set_maximum(&udp_property->nat,2);
    UDP_NatEntrySeq_set_length(&udp_property->nat,2);
#if defined(__ANDROID__)

    /* Translate the local emulator  eth0 addres 10.10.2.f:7410 to
     * 127.0.0.1:7410. This ensures that the address advertised by the
     * emulator to the host machine is the host's loopback interface, not
     * the emulator's host interface
     */
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                local_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                local_address.port = 7410;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                local_address.value.ipv4.address = 0x0a00020f;

    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                public_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                public_address.port = 7410;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                public_address.value.ipv4.address = 0x7f000001;

    /* Translate the local emulator  eth0 addres 10.10.2.f:7411 to
     * 127.0.0.1:7411. This ensures that the address advertised by the
     * emulator to the host machine is the host's loopback interface
     */
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                local_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                local_address.port = 7411;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                local_address.value.ipv4.address = 0x0a00020f;

    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                public_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                public_address.port = 7411;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                public_address.value.ipv4.address = 0x7f000001;

#else
    /* Translate the host's  lo0 addres 127.0.0.1:7412 to
     * 10.10.2.2:7412. This ensures that the address advertised by the
     * host to the emulator machine is emualtor alias to the host's loopback
     * interface.
     */
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                local_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                local_address.port = 7412;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                local_address.value.ipv4.address = 0x7f000001;

    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                public_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                public_address.port = 7412;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,0)->
                                public_address.value.ipv4.address = 0x0a000202;

    /* Translate the host's  lo0 addres 127.0.0.1:7412 to
     * 10.10.2.2:7412. This ensures that the address advertised by the
     * host to the emulator machine is emualtor alias to the host's loopback
     * interface.
     */
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                local_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                local_address.port = 7413;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                local_address.value.ipv4.address = 0x7f000001;

    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                public_address.kind = NETIO_ADDRESS_KIND_UDPv4;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                public_address.port = 7413;
    UDP_NatEntrySeq_get_reference(&udp_property->nat,1)->
                                public_address.value.ipv4.address = 0x0a000202;
#endif
#endif

    Application_printf("Register UDP");

    if (!RT_Registry_register(registry, "_udp",
                         UDP_InterfaceFactory_get_interface(),
                        (struct RT_ComponentFactoryProperty*)udp_property, NULL))
    {
        Application_printf("failed to register udp\n");
       goto done;
   }

    DDS_DomainParticipantFactory_get_qos(factory, &dpf_qos);
    dpf_qos.entity_factory.autoenable_created_entities = DDS_BOOLEAN_FALSE;
    DDS_DomainParticipantFactory_set_qos(factory, &dpf_qos);

    if (peer == NULL)
    {
        peer = "127.0.0.1"; /* default to loopback */
    }

    /* If we have a multicast address to listen on, add it. */

    /*
     * discovery_plugin_properties.multicast_receive_address = DDS_String_dup("address");
     */
    RT_Registry_register(registry,
						 "dpde",
                         DPDE_DiscoveryFactory_get_interface(),
                         &discovery_plugin_properties._parent,
                         NULL);

    OSAPI_Stdio_snprintf(dp_qos.discovery.discovery.name,8,"dpde");
    DDS_StringSeq_set_maximum(&dp_qos.discovery.initial_peers,1);
    DDS_StringSeq_set_length(&dp_qos.discovery.initial_peers,1);
    *DDS_StringSeq_get_reference(&dp_qos.discovery.initial_peers,0) = 
        DDS_String_dup(peer);

    /* if there are more remote or local endpoints, you need to increase these limits */
    dp_qos.resource_limits.max_destination_ports = 32;
    dp_qos.resource_limits.max_receive_ports = 32;
    dp_qos.resource_limits.local_topic_allocation = 1;
    dp_qos.resource_limits.local_type_allocation = 1;
    dp_qos.resource_limits.local_reader_allocation = 1;
    dp_qos.resource_limits.local_writer_allocation = 1;
    dp_qos.resource_limits.remote_participant_allocation = 8;
    dp_qos.resource_limits.remote_reader_allocation = 8;
    dp_qos.resource_limits.remote_writer_allocation = 8;

#if defined(ANDROID_EMU)
    /* With NAT we need to know what the UDP port numbers are going to be.
     * Thus, specify the participant_id manually. The android application
     * will be participant 0, the host application will be participant 1.
     */
#if defined(__ANDROID__)
    dp_qos.protocol.participant_id = 0;
#else
    dp_qos.protocol.participant_id = 1;
#endif
#endif

    application->participant =
        DDS_DomainParticipantFactory_create_participant(factory, domain_id,
                                                        &dp_qos, NULL,
                                                        DDS_STATUS_MASK_NONE);

    if (application->participant == NULL)
    {
        Application_printf("failed to create participant\n");
        goto done;
    }

    sprintf(application->type_name, "jControlAnnouncement");
    retcode = DDS_DomainParticipant_register_type(application->participant,
                                             application->type_name,
                                             HelloWorldTypePlugin_get());
    if (retcode != DDS_RETCODE_OK)
    {
        Application_printf("failed to register type: %s\n", "test_type");
        goto done;
    }

    sprintf(application->topic_name, "ControlAnnouncement");
    application->topic =
        DDS_DomainParticipant_create_topic(application->participant,
                                           application->topic_name,
                                           application->type_name,
                                           &DDS_TOPIC_QOS_DEFAULT, NULL,
                                           DDS_STATUS_MASK_NONE);

    if (application->topic == NULL)
    {
        Application_printf("topic == NULL\n");
        goto done;
    }

    success = DDS_BOOLEAN_TRUE;

  done:
    DPDE_DiscoveryPluginProperty_finalize(&discovery_plugin_properties);
    if (!success)
    {
		if (udp_property != NULL)
		{
            free(udp_property);
		}
        free(application);
        application = NULL;
    }

    return application;
}

DDS_ReturnCode_t
Application_enable(struct Application *application)
{
    DDS_Entity *entity;
    DDS_ReturnCode_t retcode;
#ifdef __ANDROID__
    const EGLint attribs[] =
    {
            EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
            EGL_BLUE_SIZE, 8,
            EGL_GREEN_SIZE, 8,
            EGL_RED_SIZE, 8,
            EGL_NONE
    };
    EGLint w, h, dummy, format;
    EGLint num_configs;
    EGLConfig config;
    EGLSurface surface;
    EGLContext context;
#endif

    entity = DDS_DomainParticipant_as_entity(application->participant);

    retcode = DDS_Entity_enable(entity);
    if (retcode != DDS_RETCODE_OK)
    {
        Application_printf("failed to enable entity\n");
        return retcode;
    }

#ifdef __ANDROID__
    Application_printf("init display\n");

#if !defined(ANDROID_EMU)
    EGLDisplay display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
    eglInitialize(display, 0, 0);
    eglChooseConfig(display, attribs, &config, 1, &num_configs);
    eglGetConfigAttrib(display, config, EGL_NATIVE_VISUAL_ID, &format);

    ANativeWindow_setBuffersGeometry(application->android_state->window,
                                     0, 0, format);

    surface = eglCreateWindowSurface(display, config,
                                     application->android_state->window, NULL);
    context = eglCreateContext(display, config, NULL, NULL);

    if (eglMakeCurrent(display, surface, surface, context) == EGL_FALSE)
    {
        Application_printf("eglMakeCurrent FAIL");
        return DDS_RETCODE_ERROR;
    }

    eglQuerySurface(display, surface, EGL_WIDTH, &w);
    eglQuerySurface(display, surface, EGL_HEIGHT, &h);

    application->display = display;
    application->context = context;
    application->surface = surface;

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    glEnable(GL_CULL_FACE);
    glShadeModel(GL_SMOOTH);
    glDisable(GL_DEPTH_TEST);
#endif
#endif
    Application_printf("Display initialized\n");

    return retcode;
}

void
Application_delete(struct Application *application)
{
    DDS_ReturnCode_t retcode;

    DDS_DomainParticipant_delete_contained_entities(application->participant);

    if (DDS_DomainParticipant_unregister_type(application->participant,
                                         application->type_name) !=
        HelloWorldTypePlugin_get())
    {
        printf("failed to unregister type: %s\n", application->type_name);
        return;
    }

    retcode =
        DDS_DomainParticipantFactory_delete_participant
        (DDS_DomainParticipantFactory_get_instance(), application->participant);
    if (retcode != DDS_RETCODE_OK)
    {
        printf("failed to delete participant: %d\n", retcode);
        return;
    }

#ifdef __ANDROID__
#if !defined(ANDROID_EMU)
    if (application->display != EGL_NO_DISPLAY)
    {
        eglMakeCurrent(application->display,
                       EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
        if (application->context != EGL_NO_CONTEXT)
        {
            eglDestroyContext(application->display, application->context);
        }
        if (application->surface != EGL_NO_SURFACE)
        {
            eglDestroySurface(application->display, application->surface);
        }
        eglTerminate(application->display);
    }

    application->display = EGL_NO_DISPLAY;
    application->context = EGL_NO_CONTEXT;
    application->surface = EGL_NO_SURFACE;
#endif
#endif

    if (application->data_sem != NULL)
    {
        OSAPI_Semaphore_delete(application->data_sem);
    }

    if (application->ready_sem != NULL)
    {
        OSAPI_Semaphore_delete(application->ready_sem);
    }

    free(application);

    DDS_DomainParticipantFactory_finalize_instance();
}

#ifdef __ANDROID__
void
Application_set_screen_color(struct Application *application,
                             DDS_Long red,DDS_Long green,DDS_Long blue)
{
#if !defined(ANDROID_EMU)
    if (application->display == NULL)
    {
        Application_printf("application->display == NULL");
        return;
    }

    /* This loosed the highest intensity color - resolution, but that is ok */
    if (red == 256)
    {
        red = 0;
    }

    if (green == 256)
    {
        green = 0;
    }

    if (blue == 256)
    {
        blue = 0;
    }
    Application_printf("set color: %p/%p,%d/%d/%d\n",
                       application->display, application->surface,
                       red,green,blue);

    glClearColor((GLfloat)red/(GLfloat)256,
                  (GLfloat)green/(GLfloat)256,
                  (GLfloat)blue/(GLfloat)256, 1);

    glClear(GL_COLOR_BUFFER_BIT);


    eglSwapBuffers(application->display, application->surface);
#endif
}
#else
void
Application_set_screen_color(struct Application *application,
                             DDS_Long red,DDS_Long green,DDS_Long blue)
{

}
#endif

