#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rti_me_c.h"
#include "wh_sm/wh_sm_history.h"
#include "rh_sm/rh_sm_history.h"

#include "HelloWorld.h"
#include "HelloWorldSupport.h"
#include "HelloWorldApplication.h"

void
HelloWorldPublisher_on_publication_matched(void *listener_data,
                                           DDS_DataWriter * writer,
                                           const struct
                                           DDS_PublicationMatchedStatus *status)
{
    Application_printf("Detected match with a subscriber\n");
}

RTI_BOOL
publisher_thread(struct OSAPI_ThreadInfo *thread_info)
{
    struct Application *application =
                                 (struct Application*)thread_info->userParam;

    DDS_Publisher *publisher;
    DDS_DataWriter *datawriter;
    HelloWorldDataWriter *hw_datawriter;
    struct DDS_DataWriterQos dw_qos = DDS_DataWriterQos_INITIALIZER;
    DDS_ReturnCode_t retcode;
    HelloWorld *sample;
    DDS_Long i;
    struct DDS_DataWriterListener dw_listener = DDS_DataWriterListener_INITIALIZER;

    sample = HelloWorld_create();
    if (sample == NULL)
    {
        Application_printf("failed to create HelloWorld sample\n");
        return RTI_FALSE;
    }

    publisher = DDS_DomainParticipant_create_publisher(application->participant,
                        &DDS_PUBLISHER_QOS_DEFAULT,NULL,DDS_STATUS_MASK_NONE);
    if (publisher == NULL)
    {
        Application_printf("publisher == NULL\n");
        goto done;
    }

#if 0
    dw_qos.reliability.kind = DDS_RELIABLE_RELIABILITY_QOS;
#else
    dw_qos.reliability.kind = DDS_BEST_EFFORT_RELIABILITY_QOS;
#endif
    dw_qos.resource_limits.max_samples = 1;
    dw_qos.resource_limits.max_samples_per_instance = 1;
    dw_qos.resource_limits.max_instances = 1;
    dw_qos.history.depth = 1;

    /* INTEROPERABILITY NOTE:
       Non-Connext Micro DDS readers will have default LivelinessQoS kind
       set to DDS_AUTOMATIC_LIVELINESS_QOS.
     
       Because Connext Micro currently only supports
       DDS_MANUAL_BY_TOPIC_LIVELINESS_QOS, the non-Connext Micro reader will need
       to set its liveliness kind to DDS_MANUAL_BY_TOPIC_LIVELINESS_QOS as well,
       in order for Requested-Offered (RxO) semantics to be compatible.
    */
 
    dw_listener.on_publication_matched =
                                    HelloWorldPublisher_on_publication_matched;

    datawriter = DDS_Publisher_create_datawriter(publisher,
       application->topic,&dw_qos,&dw_listener,DDS_PUBLICATION_MATCHED_STATUS);
    if (datawriter == NULL)
    {
        Application_printf("datawriter == NULL\n");
        goto done;
    }

    hw_datawriter = HelloWorldDataWriter_narrow(datawriter);

    if (application->wait_for_ready)
    {
        OSAPI_Semaphore_take(application->ready_sem,
                OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);
    }

    retcode = Application_enable(application);
    if (retcode != DDS_RETCODE_OK)
    {
        Application_printf("failed to enable application\n");
        goto done;
    }

    for (i = 0; (application->count > 0 && i < application->count) ||
                 (application->count == 0) && !application->exit_app; ++i)
    {

        Application_printf("sent: red=%d,green=%d,blue=%d\n",
                       sample->gl_red,sample->gl_green,sample->gl_blue);

        retcode = HelloWorldDataWriter_write(hw_datawriter,
                                             sample,&DDS_HANDLE_NIL);
        if (retcode != DDS_RETCODE_OK)
        {
            Application_printf("Failed to write to sample\n");
        }

        Application_set_screen_color(application,
                sample->gl_red,sample->gl_green,sample->gl_blue);

        if (sample->gl_blue < 256)
        {
            sample->gl_blue += 16;
        }
        else if (sample->gl_green < 256)
        {
            sample->gl_green += 16;
        }
        else if (sample->gl_red < 256)
        {
            sample->gl_red += 16;
        }
        else
        {
            sample->gl_blue = 0;
            sample->gl_green = 0;
            sample->gl_red = 0;
        }

        OSAPI_Thread_sleep(application->sleep_time);
    }

  done:

    Application_delete(application);

    DDS_DataWriterQos_finalize(&dw_qos);
    HelloWorld_finalize(sample);

    OSAPI_Semaphore_give(application->ready_sem);

    return 0;
}

#ifdef __ANDROID__
static void
publisher_handle_cmd(struct android_app* app, int32_t cmd)
{
    struct Application *application = (struct Application*)app->userData;

    switch(cmd)
    {
       case APP_CMD_INIT_WINDOW:
            if (app->window != NULL) {
                OSAPI_Semaphore_give(application->ready_sem);
            }
            break;
        case APP_CMD_TERM_WINDOW:
            break;
        default:
            break;
    }
}

static int32_t
publisher_handle_input(struct android_app* app, AInputEvent* event)
{
    return 0;
}

void
android_main(struct android_app* state)
{
    int events;
    char *argv[]={"HelloWorld_publisher"};
    int argc = 1;
    DDS_Long domain_id = 0;
#ifdef ANDROID_EMU
    char *peer = "10.0.2.2";
#else
    char *peer = "192.168.10.180";
#endif
    char *udp_intf = NULL;
    DDS_Long sleep_time = 200;
    DDS_Long count = 0;
    struct Application *application = NULL;
    struct OSAPI_Thread *pub_thread;
    struct OSAPI_ThreadProperties tprop = OSAPI_THREAD_PROPERTIES_DEFAULT;

    app_dummy();

    application = Application_create(state,
                                     "publisher", "subscriber", domain_id,
                                     udp_intf, peer, sleep_time, count);

    if (application == NULL)
    {
        return;
    }

    state->userData = application;
    state->onAppCmd = publisher_handle_cmd;
    state->onInputEvent = publisher_handle_input;

    pub_thread = OSAPI_Thread_create("pub_thread",&tprop,
                                      publisher_thread,application,NULL);
    if (pub_thread == NULL)
    {
        return;
    }

    OSAPI_Thread_start(pub_thread);

    while (1)
    {
        struct android_poll_source* source;

        while (ALooper_pollAll(-1, NULL, &events, (void**)&source) >= 0)
        {
            if (source != NULL)
            {
                source->process(state, source);
            }

            if (state->destroyRequested != 0)
            {
                Application_printf("Cleanup ....");
                application->exit_app = RTI_TRUE;
                OSAPI_Semaphore_take(application->ready_sem,
                                     OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);
                Application_delete(application);
                Application_printf("Existing");
                return;
            }
        }
    }
}
#elif !(defined(RTI_VXWORKS) && !defined(__RTP__))
int
main(int argc, char **argv)
{

    DDS_Long i = 0;
    DDS_Long domain_id = 0;
    char *peer = NULL;
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;
    struct Application *application = NULL;
    struct OSAPI_Thread *pub_thread;
    struct OSAPI_ThreadProperties tprop = OSAPI_THREAD_PROPERTIES_DEFAULT;

    for (i = 1; i < argc; ++i)
    {
        if (!strcmp(argv[i], "-domain"))
        {
            ++i;
            if (i == argc)
            {
                printf("-domain <domain_id>\n");
                return -1;
            }
            domain_id = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-udp_intf"))
        {
            ++i;
            if (i == argc)
            {
                printf("-udp_intf <interface>\n");
                return -1;
            }
            udp_intf = argv[i];
        }
        else if (!strcmp(argv[i], "-peer"))
        {
            ++i;
            if (i == argc)
            {
                printf("-peer <address>\n");
                return -1;
            }
            peer = argv[i];
        }
        else if (!strcmp(argv[i], "-sleep"))
        {
            ++i;
            if (i == argc)
            {
                printf("-sleep_time <sleep_time>\n");
                return -1;
            }
            sleep_time = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-count"))
        {
            ++i;
            if (i == argc)
            {
                printf("-count <count>\n");
                return -1;
            }
            count = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-h"))
        {
            Application_help(argv[0]);
            return 0;
        }
        else
        {
            printf("unknown option: %s\n", argv[i]);
            return -1;
        }
    }

    application = Application_create(NULL,
                                     "publisher", "subscriber", domain_id,
                                     udp_intf, peer, sleep_time, count);

    if (application == NULL)
    {
        return -1;
    }

    pub_thread = OSAPI_Thread_create("pub_thread",&tprop,
                                     publisher_thread,application,NULL);
    if (pub_thread == NULL)
    {
        return -1;
    }

    OSAPI_Thread_start(pub_thread);

    OSAPI_Semaphore_take(application->ready_sem,
            OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);

    Application_delete(application);
}
#elif defined(RTI_VXWORKS)
int
publisher_main(void)
{
    /* Explicitly configure args below */
    DDS_Long i = 0;
    DDS_Long domain_id = 0;
    char *peer = "10.10.65.104";
    char *udp_intf = NULL;
    DDS_Long sleep_time = 1000;
    DDS_Long count = 0;
    struct Application *application = NULL;
    struct OSAPI_Thread *pub_thread;
    struct OSAPI_ThreadProperties tprop = OSAPI_THREAD_PROPERTIES_DEFAULT;

    application = Application_create(NULL,
                                     "publisher", "subscriber", domain_id,
                                     udp_intf, peer, sleep_time, count);

    if (application == NULL)
    {
        return -1;
    }

    pub_thread = OSAPI_Thread_create("pub_thread",&tprop,
                                     publisher_thread,application,NULL);
    if (pub_thread == NULL)
    {
        return -1;
    }

    OSAPI_Thread_start(pub_thread);

    OSAPI_Semaphore_take(application->ready_sem,
            OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);

    Application_delete(application);
}

#endif
