#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rti_me_c.h"
#include "wh_sm/wh_sm_history.h"
#include "rh_sm/rh_sm_history.h"

#include "HelloWorld.h"
#include "HelloWorldSupport.h"
#include "HelloWorldPlugin.h"
#include "HelloWorldApplication.h"


void
HelloWorldSubscriber_on_data_available(void *listener_data,
                                       DDS_DataReader *reader)
{
    struct Application *application = (struct Application*)listener_data;
    OSAPI_Semaphore_give(application->data_sem);
}

void
HelloWorldSubscriber_read_data(void *listener_data,
                               DDS_DataReader *reader)
{
    HelloWorldDataReader *hw_reader = HelloWorldDataReader_narrow(reader);
    DDS_ReturnCode_t retcode;
    struct DDS_SampleInfo *sample_info = NULL;
    HelloWorld *sample = NULL;
    struct DDS_SampleInfoSeq info_seq =
                            DDS_SEQUENCE_INITIALIZER(struct DDS_SampleInfo);
    struct HelloWorldSeq sample_seq = DDS_SEQUENCE_INITIALIZER(HelloWorld);
    struct Application *application = (struct Application *)listener_data;
    const DDS_Long TAKE_MAX_SAMPLES = 32;
    DDS_Long i;

    retcode = HelloWorldDataReader_take(hw_reader, 
       &sample_seq, &info_seq, TAKE_MAX_SAMPLES, 
       DDS_ANY_SAMPLE_STATE, DDS_ANY_VIEW_STATE, DDS_ANY_INSTANCE_STATE);

    if (retcode == DDS_RETCODE_NO_DATA)
    {
        return;
    }

    if (retcode != DDS_RETCODE_OK)
    {
        Application_printf("failed to take data, retcode(%d)\n", retcode);
        goto done;
    }

    /* Print each valid sample taken */
    for (i = 0; i < HelloWorldSeq_get_length(&sample_seq); ++i)
    {
        sample_info = DDS_SampleInfoSeq_get_reference(&info_seq, i);
        if (sample_info->valid_data)
        {
            sample = HelloWorldSeq_get_reference(&sample_seq, i);
            Application_printf("recevied: red=%d,green=%d,blue=%d\n",
                           sample->gl_red,sample->gl_green,sample->gl_blue);
            Application_set_screen_color(application,
                    sample->gl_red,sample->gl_green,sample->gl_blue);
        }
        else
        {
            Application_printf("\nSample received\n\tINVALID DATA\n");
        }
    }

    HelloWorldDataReader_return_loan(hw_reader, &sample_seq, &info_seq);

done:
    HelloWorldSeq_finalize(&sample_seq);
    DDS_SampleInfoSeq_finalize(&info_seq);
}

void
HelloWorldSubscriber_on_subscription_matched(void *listener_data,
                                             DDS_DataReader * reader,
                                             const struct
                                             DDS_SubscriptionMatchedStatus
                                             *status)
{
    Application_printf("Detected match with a publisher\n");
}

RTI_BOOL
subscriber_thread(struct OSAPI_ThreadInfo *thread_info)
{
    struct Application *application =
                                 (struct Application*)thread_info->userParam;
    DDS_Subscriber *subscriber;
    DDS_DataReader *datareader;
    struct DDS_DataReaderQos dr_qos = DDS_DataReaderQos_INITIALIZER;
    DDS_ReturnCode_t retcode;
    struct DDS_DataReaderListener dr_listener =
                                           DDS_DataReaderListener_INITIALIZER;

    subscriber =
        DDS_DomainParticipant_create_subscriber(application->participant,
                                                &DDS_SUBSCRIBER_QOS_DEFAULT,
                                                NULL, DDS_STATUS_MASK_NONE);
    if (subscriber == NULL)
    {
        Application_printf("subscriber == NULL\n");
        goto done;
    }

    dr_listener.as_listener.listener_data = application;

#ifdef __ANDROID__
    /* The OpenGL display context is created in this thread. Thus only wake up
     * this thread when data is available so that it can update the color.
     */
    dr_listener.on_data_available = HelloWorldSubscriber_on_data_available;
#else
    dr_listener.on_data_available = HelloWorldSubscriber_read_data;
#endif
    dr_listener.on_subscription_matched =
                                  HelloWorldSubscriber_on_subscription_matched;
    dr_qos.resource_limits.max_samples = 30;
    dr_qos.resource_limits.max_instances = 10;
    dr_qos.resource_limits.max_samples_per_instance = 3;
    /* if there are more remote writers, you need to increase these limits */
    dr_qos.reader_resource_limits.max_remote_writers = 10;
    dr_qos.reader_resource_limits.max_remote_writers_per_instance = 10;
    dr_qos.history.depth = 32;

    /* Reliability QoS */
#if 0
    dr_qos.reliability.kind = DDS_RELIABLE_RELIABILITY_QOS;
#else
    dr_qos.reliability.kind = DDS_BEST_EFFORT_RELIABILITY_QOS;
#endif
    datareader = DDS_Subscriber_create_datareader(subscriber,
                          DDS_Topic_as_topicdescription(application->topic),
                          &dr_qos,&dr_listener,
                          DDS_DATA_AVAILABLE_STATUS |
                          DDS_SUBSCRIPTION_MATCHED_STATUS);

    if (datareader == NULL)
    {
        Application_printf("datareader == NULL\n");
        goto done;
    }

    /* This flag is only true for Android, and indicates that the
     * Window is ready
     */
    if (application->wait_for_ready)
    {
        Application_printf("wait for application start");
        OSAPI_Semaphore_take(application->ready_sem,
                             OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);
    }

    retcode = Application_enable(application);
    if (retcode != DDS_RETCODE_OK)
    {
        Application_printf("failed to enable application\n");
        goto done;
    }

#ifdef __ANDROID__
    /* Wait for data notification */
    while (!application->exit_app)
    {
        OSAPI_Semaphore_take(application->data_sem,
                             OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);
        HelloWorldSubscriber_read_data(application,datareader);
    }
#else
    /* Data is handled in the receive thread, the main threat just waits */
    if (application->count != 0)
    {
        Application_printf("Running for %d seconds, press Ctrl-C to exit\n",
                            application->count);
        OSAPI_Thread_sleep(application->count * 1000);
    }
    else
    {
        int sleep_loop_count =  (24 * 60 * 60) / 2000;
        int sleep_loop_left = (24 * 60 * 60) % 2000;

        Application_printf("Running for 24 hours, press Ctrl-C to exit\n");

        while (sleep_loop_count)
        {
            OSAPI_Thread_sleep(2000  * 1000); /* sleep is in ms */
            --sleep_loop_count;
        }

        OSAPI_Thread_sleep(sleep_loop_left * 1000);
    }
#endif

    done:

    DDS_DataReaderQos_finalize(&dr_qos);

    OSAPI_Semaphore_give(application->ready_sem);

    return 0;
}

#ifdef __ANDROID__
static void
subscriber_handle_cmd(struct android_app* app, int32_t cmd)
{
    struct Application *application = (struct Application*)app->userData;

    switch(cmd)
    {
       case APP_CMD_INIT_WINDOW:
            if (app->window != NULL) {
                Application_printf("start application");
                /* Tell the reader thread to start up */
                OSAPI_Semaphore_give(application->ready_sem);
            }
            break;
        case APP_CMD_TERM_WINDOW:
            /* Let the application run in the backgroud */
            break;
        default:
            break;
    }
}

static int32_t
subscriber_handle_input(struct android_app* app, AInputEvent* event)
{
    return 0;
}

void
android_main(struct android_app* state)
{
    int events;
    char *argv[]={"HelloWorld_subscriber"};
    int argc = 1;
    DDS_Long domain_id = 0;
#ifdef ANDROID_EMU
    char *peer = "10.0.2.2";
#else
    char *peer = "192.168.10.180";
#endif
    char *udp_intf = NULL;
    DDS_Long sleep_time = 200;
    DDS_Long count = 0;
    struct Application *application = NULL;
    struct OSAPI_Thread *sub_thread;
    struct OSAPI_ThreadProperties tprop = OSAPI_THREAD_PROPERTIES_DEFAULT;

    app_dummy();

    application = Application_create(state,
                                     "publisher", "subscriber", domain_id,
                                     udp_intf, peer, sleep_time, count);

    if (application == NULL)
    {
        return;
    }

    state->userData = application;
    state->onAppCmd = subscriber_handle_cmd;
    state->onInputEvent = subscriber_handle_input;

    sub_thread = OSAPI_Thread_create("sub_thread",&tprop,
                                     subscriber_thread,application,NULL);
    if (sub_thread == NULL)
    {
        return;
    }

    OSAPI_Thread_start(sub_thread);

    while (1)
    {
        struct android_poll_source* source;

        while (ALooper_pollAll(-1, NULL, &events, (void**)&source) >= 0)
        {
            if (source != NULL)
            {
                source->process(state, source);
            }

            if (state->destroyRequested != 0)
            {
                Application_printf("Cleanup ....");
                application->exit_app = RTI_TRUE;
                /* Wait for thereader thread to finish */
                OSAPI_Semaphore_take(application->ready_sem,
                                     OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);
                Application_delete(application);
                Application_printf("Existing");
                return;
            }
        }
    }
}
#elif !(defined(RTI_VXWORKS) && !defined(__RTP__))
int
main(int argc, char **argv)
{

    DDS_Long i = 0;
    DDS_Long domain_id = 0;
    char *peer = NULL;
    char *udp_intf = NULL;
    DDS_Long sleep_time = 200;
    DDS_Long count = 0;
    struct Application *application = NULL;
    struct OSAPI_Thread *sub_thread;
    struct OSAPI_ThreadProperties tprop = OSAPI_THREAD_PROPERTIES_DEFAULT;

    for (i = 1; i < argc; ++i)
    {
        if (!strcmp(argv[i], "-domain"))
        {
            ++i;
            if (i == argc)
            {
                printf("-domain <domain_id>\n");
                return -1;
            }
            domain_id = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-udp_intf"))
        {
            ++i;
            if (i == argc)
            {
                printf("-udp_intf <interface>\n");
                return -1;
            }
            udp_intf = argv[i];
        }
        else if (!strcmp(argv[i], "-peer"))
        {
            ++i;
            if (i == argc)
            {
                printf("-peer <address>\n");
                return -1;
            }
            peer = argv[i];
        }
        else if (!strcmp(argv[i], "-sleep"))
        {
            ++i;
            if (i == argc)
            {
                printf("-sleep_time <sleep_time>\n");
                return -1;
            }
            sleep_time = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-count"))
        {
            ++i;
            if (i == argc)
            {
                printf("-count <count>\n");
                return -1;
            }
            count = strtol(argv[i], NULL, 0);
        }
        else if (!strcmp(argv[i], "-h"))
        {
            Application_help(argv[0]);
            return 0;
        }
        else
        {
            printf("unknown option: %s\n", argv[i]);
            return -1;
        }
    }

    application = Application_create(NULL,
                                     "publisher", "subscriber", domain_id,
                                     udp_intf, peer, sleep_time, count);

    if (application == NULL)
    {
        return -1;
    }

    sub_thread = OSAPI_Thread_create("sub_thread",&tprop,
                                     subscriber_thread,application,NULL);
    if (sub_thread == NULL)
    {
        return -1;
    }

    OSAPI_Thread_start(sub_thread);

    OSAPI_Semaphore_take(application->ready_sem,
            OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);

    Application_delete(application);
}
#elif defined(RTI_VXWORKS)
int
publisher_main(void)
{
    /* Explicitly configure args below */
    DDS_Long i = 0;
    DDS_Long domain_id = 0;
    char *peer = "10.10.65.104";
    char *udp_intf = NULL;
    DDS_Long sleep_time = 200;
    DDS_Long count = 0;
    struct Application *application = NULL;
    struct OSAPI_Thread *sub_thread;
    struct OSAPI_ThreadProperties tprop = OSAPI_THREAD_PROPERTIES_DEFAULT;

    application = Application_create(NULL,
                                     "publisher", "subscriber", domain_id,
                                     udp_intf, peer, sleep_time, count);

    if (application == NULL)
    {
        return -1;
    }

    sub_thread = OSAPI_Thread_create("sub_thread",&tprop,
                                     subscriber_thread,application,NULL);
    if (sub_thread == NULL)
    {
        return -1;
    }

    OSAPI_Thread_start(sub_thread);

    OSAPI_Semaphore_take(application->ready_sem,
            OSAPI_SEMAPHORE_TIMEOUT_INFINITE,NULL);

    Application_delete(application);
}
#endif
